#version 330 core

uniform mat3 nMatrix;

uniform sampler2D textureMap;
uniform sampler2D textureMap2; // height map
uniform sampler2D textureMap3; // normal map

uniform float diffusePower;
uniform float specularPower;
uniform float ambientPower;
uniform float specularReflectionExponent; // e.g. 5
uniform float textureScale1; // height map scale factor, e.g. 300
uniform float transparency;

in vec3 viewDirection_eye;
in vec3 lightDirection_eye;
in vec2 uv;
in vec2 posXZ_world;

out vec4 color;

void main(void)
{
	vec2 texCoordsToUse = vec2(uv.x, -uv.y);

	// --- calculate normals:
	// 1. get values (= heights) of surrounding vertices:
	float left   = texture(textureMap2, (posXZ_world + vec2(-1,  0)) / textureScale1).x;
	float right  = texture(textureMap2, (posXZ_world + vec2( 1,  0)) / textureScale1).x;
	float top    = texture(textureMap2, (posXZ_world + vec2( 0, -1)) / textureScale1).x;
	float bottom = texture(textureMap2, (posXZ_world + vec2( 0,  1)) / textureScale1).x;

	// 2. calculate tangent vectors:
	vec3 t1 = normalize(vec3(1.0, right - left, 0.0));
	vec3 t2 = normalize(vec3(0.0, bottom - top, -1.0));

	// 3. calculate normal:
	vec3 n = cross(t1, t2);
	vec3 normal_eye = nMatrix * n;
	normal_eye = normalize(normal_eye);
	
	// convert tangents to eye space:
	vec3 tangent_eye = normalize(nMatrix * t1);
	vec3 bitangent_eye = normalize(nMatrix * t2);

	// switch to tangent space:
	mat3 tangentMatrix = mat3(tangent_eye, bitangent_eye, normal_eye);
	mat3 invTangentMatrix = transpose(tangentMatrix);

	vec3 viewDirection_tangent = invTangentMatrix * viewDirection_eye;
	viewDirection_tangent = normalize(viewDirection_tangent); // necessary?

	vec3 lightDirection_tangent = invTangentMatrix * lightDirection_eye;
	lightDirection_tangent = normalize(lightDirection_tangent); // necessary?

	vec3 normal_tangent = texture2D(textureMap3, texCoordsToUse).xyz; // sample from normal map
	normal_tangent = normalize(normal_tangent * 2.0 - 1.0); // convert from [0,1] to [-1,1] range


	// FOR TEST PURPOSE ONLY: (Turn normal mapping off)
	//lightDirection_tangent = lightDirection_eye;
	//viewDirection_tangent = viewDirection_eye;
	//normal_tangent = normal_eye;
			

	// Blinn Phong shading:

	vec3 L = lightDirection_tangent;
	vec3 N = normal_tangent;
	vec3 V = viewDirection_tangent;

	vec4 materialDiffuseColor = texture2D(textureMap, texCoordsToUse);
	vec4 materialAmbientColor = ambientPower * materialDiffuseColor;
	vec4 materialSpecularColor = vec4(0.3);

	float distance = length(L);
	L = L / distance; // = normalize(L)
	distance = distance * distance;

	float intensityDiffuse = clamp(dot(N,L), 0, 1);
	//vec4 diffuseColor = intensityDiffuse * materialDiffuseColor * diffusePower / distance;
	vec4 diffuseColor = intensityDiffuse * materialDiffuseColor * diffusePower;

	vec3 H = normalize(L + V); // half vector

	float intensitySpecular = clamp(dot(N, H), 0, 1);
	intensitySpecular = pow(intensitySpecular, specularReflectionExponent);
	//vec4 specularColor = intensitySpecular * materialSpecularColor * specularPower / distance;
	vec4 specularColor = intensitySpecular * materialSpecularColor * specularPower;

	vec4 ambientColor = materialAmbientColor;
	color = vec4((ambientColor + diffuseColor + specularColor).xyz, transparency);
}
